/* $Id: shed.cpp,v 1.22 2002/09/12 16:08:10 alexsisson Exp $
 *
 * shed source code
 *
 * (c) Alex Sisson 2002
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* includes */
#include <stdio.h>     /* fprintf    */
#include <strstream.h> /* searching  */
#include <fstream.h>   /* fstream    */
#include <ncurses.h>   /* ncurses    */
#include <signal.h>    /* signal     */
#include <string.h>    /* strcmp etc */
#include <stdlib.h>    /* exit       */
#include <sys/types.h> /* stat       */
#include <sys/stat.h>  /*  "         */
#include <unistd.h>    /*  "         */
#include <math.h>      /* pow,log    */

/* defines */
#define SHEDVERSION    "1.08"

#define strequ(s,t)    (!strcmp(s,t))

#ifdef SHEDDEBUG
#define debug(str)     fprintf(stderr,"debug: %s",str);
#else
#define debug(str)
#endif

/* typedefs */
typedef unsigned char  byte;
typedef unsigned short word;
typedef unsigned long  dword;

/* function prototypes */
static void finish(int sig);
char*  getstring(dword n, char* s, word base = 10, word width = 0);
long   parsestring(char* s, word base);
word   calcwidth(dword n, word base = 10);
void   cursorup();
void   cursorup(dword n);
void   cursordown();
void   cursordown(dword n);
void   cursorleft();
void   cursorright();
void   clearline(word line, word width = COLS);
bool   search(char* str);

/* globals */
fstream f;
int     len           = 0;                               /* file length */
int     cursorx       = 0;                               /* offset of cursor from SOF */
int     cursory       = 0;                               /* which column cursor is in */
int     viewport      = 0;                               /* offset of current view from SOF */
int     viewsize      = 0;                               /* size of viewport */
bool    decmode       = true;                            /* dec or hex display */
bool    readonly      = false;                           /* readonly file */
dword   offsetwidth   = 0;                               /* width of left column */
word    colbase[5]    = {0,16,10,8,2};                   /* base of each column */
word    colwidth[5]   = {1,2,3,3,8};                     /* width of columns */
word    coloffset[5]  = {0,6,10,14,18};                  /* offset from left for each column */
char*   coltitle[5]   = {"asc","hex","dec","oct","bin"}; /* name of each column */
char*   searchstr[5]  = {NULL,NULL,NULL,NULL,NULL};      /* previous searches */

/* main */
int main(int argc, char* argv[])
{
  /* process args */
  if(!argv[1])
  {
    fprintf(stderr,"%s: no file specified\n",argv[0]);
    return 1;
  }
  if(strequ(argv[1],"-h")||strequ(argv[1],"--help"))
  {
    fprintf(stdout,"usage: %s <filename>\n",argv[0]);
    return 0;
  }
  if(strequ(argv[1],"-v")||strequ(argv[1],"--version"))
  {
    fprintf(stdout,"%s %s\n",argv[0],SHEDVERSION);
    return 0;
  }

  /* open file */
  char* filename = argv[1];
  f.open(filename,ios::in|ios::out|ios::binary|ios::nocreate);
  if(!f)
  {
    /* try to open readonly */
    f.open(filename,ios::in|ios::binary);
    if(!f)
    {
      fprintf(stderr,"%s: could not open %s\n",argv[0],filename);
      return 2;
    }
    readonly = true;
  }

  /* get file stats */
  struct stat stats;
  if(stat(argv[1],&stats)<0)
  {
    fprintf(stderr,"%s: could not stat %s - perhaps it is too big (>2Gb)\n",argv[0],argv[1]);
    return 1;
  }

  if(S_ISDIR(stats.st_mode))
  {
    fprintf(stderr,"%s: %s is a directory\n",argv[0],argv[1]);
    return 1;
  }

  len = stats.st_size;

  /* init ncurses */
  signal(SIGINT,finish);
  initscr();
  keypad(stdscr,TRUE);
  cbreak();
  noecho();
  halfdelay(10);

  /* set size of viewport to LINES - 6 (2 reservered for top + 4 for bottom area) */
  viewsize = LINES - 6;

  /* calculate the width for the offset column and round it */
  offsetwidth = calcwidth(len);
  if(!len) offsetwidth = 16;
  else if(offsetwidth<=4) offsetwidth = 4;
  else if(offsetwidth<=8) offsetwidth = 8;
  else if(offsetwidth<=12) offsetwidth = 12;
  else if(offsetwidth<=16) offsetwidth = 16;

  char* str = new char[256];
  long i;
  byte c;
  ofstream of;

  /* main loop */
  while(1)
  {
    /* redraw top */
    attron(A_REVERSE);
    clearline(0);
    mvaddstr(0,0,"shed ");
    addstr(SHEDVERSION);
    addstr("  file: ");
    addstr(filename);
    if(readonly) addstr(" (readonly)");
    attroff(A_REVERSE);

    /* draw column headers */
    mvaddstr(1,0,(offsetwidth==4)?"offs":"offset");
    mvaddstr(1,offsetwidth+2,"asc hex dec oct bin");

    /* seek to current part of file and display */
    f.clear();
    f.seekg(viewport,ios::beg);
    for(i=0;i<viewsize;i++)
    {
      f.get(c);
      if(f.fail()) break;
      move(i+2,0);
      addstr(getstring(viewport+i,str,(decmode)?10:16,offsetwidth));
      addstr(":  ");
      addch((c>32&&c<127)?c:' ');
      addstr("  ");
      addstr(getstring(c,str,16,2));
      addstr("  ");
      addstr(getstring(c,str,10,3));
      addstr(" ");
      addstr(getstring(c,str, 8,3));
      addstr(" ");
      addstr(getstring(c,str, 2,8));
    }

    /* draw (end) if necessary */
    mvaddstr(i+1,offsetwidth + coloffset[4] + colwidth[4] + 1,(viewport+i==len)?"(end)":"     ");

    /* draw cursor */
    f.clear();
    f.seekg(cursorx,ios::beg);
    f.get(c);
    attron(A_REVERSE);
    dword pos = (cursorx - viewport) + 2;
    if(!cursory) mvaddch(pos,offsetwidth+3,(c>32&&c<127)?c:' ');
    else mvaddstr(pos,offsetwidth+coloffset[cursory],getstring(c,str,colbase[cursory],colwidth[cursory]));

    /* draw cursor pos */
    clearline(LINES-3);
    c = (decmode)?10:16;
    mvaddstr(LINES-3,COLS-(calcwidth(cursorx,c)+calcwidth(len,c)+8),getstring(cursorx,str,c));
    addstr("/");
    addstr(getstring(len,str,c));
    addstr((decmode)?" (dec)":" (hex)");

    attroff(A_REVERSE);

    /* draw key help */
    clearline(LINES-2);
    clearline(LINES-1);
    mvaddstr(LINES-2,0,"[SPACE/E] edit  [S/W/F] search         [J] jump to       [T] toggle dec/hex");
    mvaddstr(LINES-1,0,"[X]       exit  [R]     repeat search  [D] dump to file");
    mvaddstr(LINES-1,COLS-1," ");
    refresh();

    /* process keypress */
    switch(getch())
    {
      case KEY_UP:
        cursorup();
        break;

      case KEY_DOWN:
        cursordown();
        break;

      case KEY_LEFT:
        cursorleft();
        break;

      case KEY_RIGHT:
      case 9: /* tab */
        cursorright();
        break;

      case KEY_PPAGE:
      case 25: /* ctrl Y */
        cursorup(16);
        break;

      case KEY_NPAGE:
      case 22: /* ctrl V */
        cursordown(16);
        break;

      case KEY_HOME:
      case 1: /* ctrl A */
        cursory = 0;
        break;

      case KEY_END:
      case 5: /* ctrl E */
        cursory = 4;
        break;

      /* edit */
      case ' ':
      case 'E':
      case 'e':
        attron(A_REVERSE);
        clearline(LINES-3,24);
        if(readonly)
        {
          beep();
          break;
        }
        mvaddstr(LINES-3,0,"new value (");
        f.seekp(cursorx,ios::beg);
        addstr(coltitle[cursory]);
        addstr("): ");
        echo();
        getstr(str);
        noecho();
        attroff(A_REVERSE);
        if(!strlen(str)) break;
        if(!cursory) f.put((byte)str[0]);
        else
        {
          long l = parsestring(str,colbase[cursory]);
          if(l<0 || l>255) break;
          f.put((byte)l);
        }
        break;

      /* exit */
      case 'X':
      case 'x':
      case 24: //ctrl X
        finish(0);
        break;

      /* jump to */
      case 'J':
      case 'j':
        attron(A_REVERSE);
        clearline(LINES-3,24);
        mvaddstr(LINES-3,0,"jump to ");
        addstr((decmode)?"(dec): ":"(hex): ");
        echo();
        getstr(str);
        noecho();
        attroff(A_REVERSE);
        if(!strlen(str)) break;
        long target;
        if(strequ(str,"top")) target = 0;
        else if(strequ(str,"end")) target = len-1;
        else
        {
          target = parsestring(str,(decmode)?10:16);
          if(target<0) break;
          if(target>=len) target = len-1;
        }
        if(cursorx>target)
        {
          if(target<viewport || target>viewport+viewsize)
          {
            cursorx = target;
            viewport = cursorx;
          }
          else cursorup(cursorx-target);
        }
        else
        {
          if(target<viewport || target>viewport+viewsize)
          {
            cursorx = target;
            viewport = cursorx - (viewsize - 1);
          }
          else cursordown(target-cursorx);
        }
        break;

      /* search */
      case 'S':
      case 's':
      case 'W':
      case 'w':
      case 'F':
      case 'f':
        attron(A_REVERSE);
        clearline(LINES-3,24);
        mvaddstr(LINES-3,0,"search for (");
        addstr(coltitle[cursory]);
        addstr(")");
        if(searchstr[cursory])
        {
          addstr("[");
          addstr(searchstr[cursory]);
          addstr("]");
        }
        addstr(": ");
        echo();
        getstr(str);
        noecho();
        attroff(A_REVERSE);
        search(str);
        break;

      /* repeat search */
      case 'R':
      case 'r':
        if(searchstr[cursory]) search("");
        break;

      /* toggle */
      case 'T':
      case 't':
        decmode = !decmode;
        break;

      /* dump to file */
      case 'D':
      case 'd':
        attron(A_REVERSE);
        clearline(LINES-3,24);
        mvaddstr(LINES-3,0,"dump to file: ");
        echo();
        getstr(str);
        noecho();
        attroff(A_REVERSE);
        if(!strlen(str)) break;
        of.open(str,ios::out|ios::binary);
        if(!of) break;
        of << ((offsetwidth==4)?"offs":"offset") << "    asc hex dec oct bin\n";
        f.clear();
        f.seekg(0,ios::beg);
        for(i=0;i<len;i++)
        {
          f.get(c);
          if(f.fail())
          {
            break;
          }
          of << getstring(viewport+i,str,(decmode)?10:16,offsetwidth);
          of << ":  ";
          of << (char)((c>32&&c<127)?c:' ') << "  ";
          of << getstring(c,str,16,2) << "  ";
          of << getstring(c,str,10,3) << " ";
          of << getstring(c,str, 8,3) << " ";
          of << getstring(c,str, 2,8) << "\n";
        }
        of.close();
        break;

      /* redraw */
      case 12: //^L
        erase();
        refresh();
        break;

      /* resize */
      case KEY_RESIZE:
        refresh();
        viewsize = LINES - 6;
        if(cursorx>=viewport+viewsize)
        {
          viewport = cursorx;
        }
        if(viewport+viewsize>len)
        {
          while(viewport+viewsize>len)
          {
            cursorup();
          }
          cursordown(viewsize);
        }
        for(i=1;i<(word)(LINES-1);i++)
        {
          clearline(i);
        }
        refresh();
        break;

      default:
        break;
    }
  }
  finish(0);
}

/* functions */

/* ends ncurses and quits */
static void finish(int sig)
{
  endwin();
  cout << "\n";
  exit(sig);
}

/* returns string representation of n in base 'base' */
char* getstring(dword n, char* s, word base = 10, word width = 0)
{
  dword i,j,c,d = calcwidth(n,base);
  char* p = s;
  if(width && d<width)
  {
    for(i=d;i<width;i++)
    {
      *p = '0';
      p++;
    }      
  }
  for(i=d;i;i--)
  {
    j = (dword)pow(base,i-1);
    c = 0;
    while(n>=j)
    {
      n-=j;
      c++;
    }
    *p = c + ((c<10) ? 48 : 55);
    p++;
  }
  *p = 0;
  return s;
}

/* parses 's' as a 'base' base string */
long parsestring(char* s, word base)
{
  debug("parseing " << s << " as base " << base << "\n");
  word i,j,slen=strlen(s);
  dword r=0,n=0;

  for(i=0;i<slen;i++)
  {
    if(s[i]>='0' && s[i]<='9')      n = s[i] - 48;
    else if(s[i]>='A' && s[i]<='F') n = s[i] - 55;
    else if(s[i]>='a' && s[i]<='f') n = s[i] - 87;
    else return -1;
    if(n>=base) return -1;
    for(j=1;j<slen-i;j++) n*=base;
    r+=n;
  }
  debug("result = " << r << "\n");
  return r;
}

/* calculates the width needed to represent 'n' in base 'base' */
word calcwidth(dword n, word base = 10)
{
  dword i,j,b;
  for(i=0;;i++)
  {
    b = base;
    for(j=0;j<i;j++) b *= base;
    if(b>n || b<base) return i + 1;
  }
  return 0;
}

/* cursor movements functions */
void cursorup()
{
  if(cursorx)
  {
    cursorx--;
    if(cursorx<viewport) viewport--;
  }
  else beep();
}

void cursordown()
{
  if(cursorx<len-1)
  {
    cursorx++;
    if(cursorx>=viewport+viewsize) viewport++;
  }
  else beep();
}

void cursorleft()
{
  if(cursory) cursory--;
  else beep();
}

void cursorright()
{
  if(cursory<4) cursory++;
  else beep();
}

void cursorup(dword n)
{
  while(n--) cursorup();
}

void cursordown(dword n)
{
  while(n--) cursordown();
}

/* clears a line on the screen */
void clearline(word line, word width = COLS)
{
  move(line,0);
  for(line=0;line<width;line++) addch(' ');
}

/* search */
bool search(char* str)
{
  if(strlen((char*)str))
  {
    //user enter a string, so make a copy to searchstr for repeat searches.
    if(searchstr[cursory]) delete searchstr[cursory];
    searchstr[cursory] = new char[strlen(str)+1];
    strcpy(searchstr[cursory],str);
  }
  else if(!searchstr[cursory])
  {
    //else user just pressed enter, but no previous search
    return false;
  }

  byte c;
  word i,slen = strlen(searchstr[cursory]);
  byte* search = new byte[slen+2];
  strcpy((char*)search,searchstr[cursory]);
  long l;

  if(cursory)
  {
    char* p = new char[slen+2];
    strcpy(p,(char*)search);
    p = strtok(p," :,.\0");
    for(i=0;p;i++)
    {
      l = parsestring(p,colbase[cursory]);
      if(l<0 || l>255) return false;
      search[i] = (byte)l;
      p = strtok(NULL," :,.\0");
    }
    delete p;
    search[i] = 0;
    slen = i;
    debug("searching for " << str << "\n");
  }

  cursordown();

  f.clear();
  f.seekg(cursorx,ios::beg);
  for(;cursorx<len-1;cursordown())
  {
    debug("offset = " << cursorx << "\n");
    f.get(c);
    debug("c = " << (char)((c>=32&&c<=120)?c:' ') << ":" << (int)c << "\n");
    if(c==search[0])
    {
      debug("start match\n");
      for(i=1;i<slen;i++)
      {
        f.get(c);
        if(c!=search[i]) break;
      }
      if(i==slen)
      {
        cursordown(i-1);
        cursorup(i-1);
        return true;
      }
      f.clear();
      f.seekg(cursorx+1,ios::beg);
    }
  }
  return false;
}
